/*
 * Project: MoleCuilder
 * Description: creates and alters molecular systems
 * Copyright (C)  2010-2012 University of Bonn. All rights reserved.
 * 
 *
 *   This file is part of MoleCuilder.
 *
 *    MoleCuilder is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    MoleCuilder is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with MoleCuilder.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * AdjacencyListUnitTest.cpp
 *
 *  Created on: Oct 17, 2011
 *      Author: heber
 */

// include config.h
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "AdjacencyListUnitTest.hpp"

#include <cppunit/CompilerOutputter.h>
#include <cppunit/extensions/TestFactoryRegistry.h>
#include <cppunit/ui/text/TestRunner.h>

#include <boost/lambda/lambda.hpp>
#include <iostream>
#include <iterator>
#include <iostream>
#include <vector>

#include "CodePatterns/Assert.hpp"
#include "CodePatterns/Log.hpp"

#include "Atom/atom.hpp"
#include "Descriptors/AtomDescriptor.hpp"
#include "Element/element.hpp"
#include "Element/periodentafel.hpp"
#include "Graph/AdjacencyList.hpp"
#include "molecule.hpp"
#include "World.hpp"
#include "WorldTime.hpp"

#ifdef HAVE_TESTRUNNER
#include "UnitTestMain.hpp"
#endif /*HAVE_TESTRUNNER*/

/********************************************** Test classes **************************************/
// Registers the fixture into the 'registry'
CPPUNIT_TEST_SUITE_REGISTRATION( AdjacencyListTest );

static std::string adjacencyfile ="\
1 2\n\
2 1 3\n\
3 2 4\n\
4 3 5\n\
5 4 6\n\
6 5 7\n\
7 6 8\n\
8 7 9\n\
9 8 10\n\
10 9\n";

static std::string wrongadjacencyfile1 ="\
1 2\n\
2 1\n\
4 5\n\
5 4 6\n\
6 5 7\n\
7 6 8\n\
8 7 9\n\
9 8 10\n\
10 9\n";

static std::string wrongadjacencyfile2 ="\
1 2\n\
2 1 3\n\
3 2 4\n\
4 3 5\n\
5 4 6\n\
6 5 7\n\
7 6 8\n\
8 7 9\n\
9 8 10\n\
10 9 11\n\
11 10";

static std::string wrongadjacencyfile3 ="\
1 2\n\
2 1 3\n\
3 2 4\n\
4 3 5\n\
5 4 7\n\
6\n\
7 5 8\n\
8 7 9\n\
9 8 10\n\
10 9 11\n\
11 10";

// set up and tear down
void AdjacencyListTest::setUp()
{
  // failing asserts should be thrown
  ASSERT_DO(Assert::Throw);

  const element *hydrogen = World::getInstance().getPeriode()->FindElement(1);
  CPPUNIT_ASSERT(hydrogen != NULL);

  TestMolecule = World::getInstance().createMolecule();
  CPPUNIT_ASSERT(TestMolecule != NULL);
  for(int i=0;i<ATOM_COUNT;++i){
    atom *_atom = World::getInstance().createAtom();
    CPPUNIT_ASSERT(_atom != NULL);
    _atom->setType(hydrogen);
    TestMolecule->AddAtom(_atom);
    atoms.push_back(_atom);
    atomIds.push_back(_atom->getId());
  }
  CPPUNIT_ASSERT_EQUAL( (size_t)ATOM_COUNT, atoms.size());
  CPPUNIT_ASSERT_EQUAL( (size_t)ATOM_COUNT, atomIds.size());
  // create linear chain
  for(int i=0;i<ATOM_COUNT-1;++i)
    atoms[i]->addBond(WorldTime::getTime(), atoms[i+1]);

  // create map as it should be
  for(int i=0;i<ATOM_COUNT;++i) {
    if (i != 0) // first has only one bond
      comparisonMap.insert( std::make_pair(atomIds[i], atomIds[i-1]) );
    if (i != ATOM_COUNT-1) // last has only one bond
      comparisonMap.insert( std::make_pair(atomIds[i], atomIds[i+1]) );
  }
}

void AdjacencyListTest::tearDown()
{
  comparisonMap.clear();

  // destroy molecule and contained atoms
  TestMolecule->removeAtomsinMolecule();
  World::getInstance().destroyMolecule(TestMolecule);
  // destroy World
  World::purgeInstance();
//  logger::purgeInstance();
//  errorLogger::purgeInstance();
}

/** Unit tests for AdjacencyList::CreateExternalMap().
 *
 */
void AdjacencyListTest::CreateExternalMapTest()
{
  std::stringstream input;
  AdjacencyList fileChecker(input);
  fileChecker.CreateExternalMap(atomIds);

  // check size (it's 8*2 + 2*1 = 18 keys)
  CPPUNIT_ASSERT_EQUAL( (size_t)18, fileChecker.ExternalAtomBondMap.size() );
  CPPUNIT_ASSERT_EQUAL( (size_t)0, fileChecker.InternalAtomBondMap.size() );

  // check equality
  CPPUNIT_ASSERT( comparisonMap.size() == fileChecker.ExternalAtomBondMap.size() );
//  std::cout << "comparisonMap: " << comparisonMap << std::endl;
//  std::cout << "fileChecker.InternalAtomBondMap: " << fileChecker.InternalAtomBondMap << std::endl;
  CPPUNIT_ASSERT( comparisonMap == fileChecker.ExternalAtomBondMap );

  // check non-equality: more
  comparisonMap.insert( std::make_pair( (atomId_t)10, (atomId_t)8) );
  CPPUNIT_ASSERT( comparisonMap != fileChecker.ExternalAtomBondMap );
  comparisonMap.erase((atomId_t)10);

  // check non-equality: less
  comparisonMap.erase((atomId_t)9);
  CPPUNIT_ASSERT( comparisonMap != fileChecker.ExternalAtomBondMap );
}

/** Unit tests for AdjacencyList::ParseInInternalMap().
 *
 */
void AdjacencyListTest::ParseInInternalMapTest()
{
  std::stringstream input(adjacencyfile);
  AdjacencyList fileChecker(input);
  std::vector<atomId_t> noids;
  fileChecker.CreateExternalMap(noids);

  // check size (it's 8*2 + 2*1 = 18 keys)
  CPPUNIT_ASSERT_EQUAL( (size_t)0, fileChecker.ExternalAtomBondMap.size() );
  CPPUNIT_ASSERT_EQUAL( (size_t)18, fileChecker.InternalAtomBondMap.size() );

  // check equality
  CPPUNIT_ASSERT( comparisonMap.size() == fileChecker.InternalAtomBondMap.size() );
  CPPUNIT_ASSERT( comparisonMap == fileChecker.InternalAtomBondMap );

  // check non-equality: more
  comparisonMap.insert( std::make_pair( (atomId_t)10, (atomId_t)8) );
  CPPUNIT_ASSERT( comparisonMap != fileChecker.InternalAtomBondMap );
  comparisonMap.erase((atomId_t)10);

  // check non-equality: less
  comparisonMap.erase((atomId_t)9);
  CPPUNIT_ASSERT( comparisonMap != fileChecker.InternalAtomBondMap );
}

/** Unit tests for AdjacencyList::CompareInternalExternalMap().
 *
 */
void AdjacencyListTest::CompareInternalExternalMapTest()
{
  std::stringstream input(adjacencyfile);
  AdjacencyList fileChecker(input);

  // assert equality before parsing (empty sets should always return true)
  CPPUNIT_ASSERT( fileChecker.ExternalAtomBondMap.size() != fileChecker.InternalAtomBondMap.size() );
  CPPUNIT_ASSERT( fileChecker.ExternalAtomBondMap != fileChecker.InternalAtomBondMap );
  CPPUNIT_ASSERT( fileChecker.CompareInternalExternalMap() );

  // parse in external map
  fileChecker.CreateExternalMap(atomIds);

  // assert equality after parsing
  CPPUNIT_ASSERT_EQUAL( fileChecker.ExternalAtomBondMap.size(), fileChecker.InternalAtomBondMap.size() );
  CPPUNIT_ASSERT_EQUAL( fileChecker.ExternalAtomBondMap, fileChecker.InternalAtomBondMap );
  CPPUNIT_ASSERT( fileChecker.CompareInternalExternalMap() );
}

/** Unit tests for AdjacencyList::operator()().
 *
 */
void AdjacencyListTest::CheckAgainstSubsetTest()
{
  {
    // parse right
    std::stringstream input(adjacencyfile);
    AdjacencyList fileChecker(input);
    CPPUNIT_ASSERT( fileChecker.CheckAgainstSubset(atomIds) );
  }
  {
    // parse wrong1 (more atoms in the world than in file, hence wrong)
    std::stringstream input(wrongadjacencyfile1);
    AdjacencyList fileChecker(input);
    CPPUNIT_ASSERT( !fileChecker.CheckAgainstSubset(atomIds) );
  }
  {
    // remove third atom (index starts at 0) and test for equality then
    std::vector<atomId_t> validids;
    std::remove_copy_if(atomIds.begin(), atomIds.end(), std::back_inserter(validids), boost::lambda::_1 == (atomId_t)2);
    CPPUNIT_ASSERT_EQUAL( (size_t)ATOM_COUNT-1, validids.size() );
    // parse wrong1 (more atoms in the world than in file, hence wrong)
    std::stringstream input(wrongadjacencyfile1);
    AdjacencyList fileChecker(input);
    CPPUNIT_ASSERT( fileChecker.CheckAgainstSubset(validids) );
  }
  {
    // parse wrong2 (there is no atom 10, but present in file. hence true)
    std::stringstream input(wrongadjacencyfile2);
    AdjacencyList fileChecker(input);
    CPPUNIT_ASSERT( !fileChecker.CheckAgainstSubset(atomIds) );
  }
  {
    // parse wrong3 (6 is not connected)
    std::stringstream input(wrongadjacencyfile3);
    AdjacencyList fileChecker(input);
    CPPUNIT_ASSERT( !fileChecker.CheckAgainstSubset(atomIds) );
  }
}
